#!/usr/bin/env python

"""
clana is a toolkit for classifier analysis.

It specifies some file formats and comes with some tools for typical tasks of
classifier analysis.
"""
# core modules
import logging.config
import os
import random

# 3rd party modules
import click
import matplotlib

matplotlib.use("Agg")  # noqa

# internal modules
import clana
import clana.visualize_cm
import clana.get_cm
import clana.get_cm_simple
import clana.distribution

config = clana.utils.load_cfg(verbose=True)
logging.config.dictConfig(config["LOGGING"])
logging.getLogger("matplotlib").setLevel("WARN")
random.seed(0)


@click.group()
@click.version_option(version=clana.__version__)
def entry_point():
    pass


@entry_point.command(name="get-cm-simple")
@click.option(
    "--labels",
    "label_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
@click.option(
    "--gt",
    "gt_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
@click.option(
    "--predictions",
    "predictions_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
@click.option(
    "--clean",
    default=False,
    is_flag=True,
    help="Remove classes that the classifier doesn't know",
)
def get_cm_simple(label_filepath, gt_filepath, predictions_filepath, clean):
    clana.get_cm_simple.main(label_filepath, gt_filepath, predictions_filepath, clean)


@entry_point.command(name="get-cm")
@click.option(
    "--predictions",
    "cm_dump_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
@click.option(
    "--gt",
    "gt_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
@click.option("--n", "n", required=True, type=int, help="Number of classes")
def get_cm(cm_dump_filepath, gt_filepath, n):
    """Generate a confusion matrix from predictions and ground truth."""
    clana.get_cm.main(cm_dump_filepath, gt_filepath, n)


@entry_point.command(name="distribution")
@click.option(
    "--gt",
    "gt_filepath",
    required=True,
    type=click.Path(exists=True),
    help="List of labels for the dataset",
)
def distribution(gt_filepath):
    """Get the distribution of classes in a dataset."""
    clana.distribution.main(gt_filepath)


@entry_point.command(name="visualize")
@click.option("--cm", "cm_file", type=click.Path(exists=True), required=True)
@click.option(
    "--perm",
    "perm_file",
    help="json file which defines a permutation to start with.",
    type=click.Path(),
    default=None,
)
@click.option(
    "--steps",
    default=1000,
    show_default=True,
    help="Number of steps to find a good permutation.",
)
@click.option("--labels", "labels_file", default="")
@click.option(
    "--zero_diagonal",
    is_flag=True,
    help=(
        "Good classifiers have the highest elements on the diagonal. "
        "This option sets the diagonal to zero so that errors "
        "can be seen more easily."
    ),
)
@click.option(
    "--output",
    "output_image_path",
    type=click.Path(exists=False),
    help="Where to store the image (either .png or .pdf)",
    default=os.path.abspath(config["visualize"]["save_path"]),
    show_default=True,
)
@click.option(
    "--limit_classes", type=int, help="Limit the number of classes in the output"
)
def visualize(
    cm_file,
    perm_file,
    steps,
    labels_file,
    zero_diagonal,
    output_image_path,
    limit_classes=None,
):
    """Optimize and visualize a confusion matrix."""
    print_file_format_issues(cm_file, labels_file, perm_file)
    clana.visualize_cm.main(
        cm_file,
        perm_file,
        steps,
        labels_file,
        zero_diagonal,
        limit_classes,
        output_image_path,
    )


def print_file_format_issues(cm_file, labels_file, perm_file):
    """
    Get all potential issues of the file formats.

    Parameters
    ----------
    cm_file : str
    labels_file : str
    perm_file : str
    """
    if not (cm_file.lower().endswith("json") or cm_file.lower().endswith("csv")):
        print(
            "[WARNING] A json file is expected for the cm_file, but was {}".format(
                cm_file
            )
        )
    if not (perm_file is None or perm_file.lower().endswith("json")):
        print(
            "[WARNING] A json file is expected fo the perm_file, but was {}".format(
                perm_file
            )
        )
    cm = clana.io.read_confusion_matrix(cm_file)
    labels = clana.io.read_labels(labels_file, len(cm))
    special_labels = ["UNK"]
    if len(labels) - len(special_labels) < len(cm):
        print(
            "[WARNING] The shape of the confusion matrix is {cm_shape}, but "
            "only {nb_labels} labels were found: {labels}".format(
                cm_shape=cm.shape, nb_labels=len(labels), labels=labels
            )
        )
        print(
            "Please keep in mind that the first row of the labels file is "
            "the header of the CSV (delimiter: ;)"
        )



#!/usr/bin/env python

"""Get the distribution of classes in a dataset."""


def main(gt_filepath):
    """
    Get the distribution of classes in a file.

    Parameters
    ----------
    gt_filepath : str
        List of ground truth; one label per line
    """
    # Read text file
    with open(gt_filepath, "r") as fp:
        read_lines = fp.readlines()
        labels = [line.rstrip("\n") for line in read_lines]

    distribution = get_distribution(labels)
    labels = sorted(distribution.items(), key=lambda n: (-n[1], n[0]))
    label_len = max(len(label[0]) for label in labels)
    count_len = max(len(str(label[1])) for label in labels)
    total_count = sum(label[1] for label in labels)
    for label, count in labels:
        print(
            "{percentage:5.2f}% {label:<{label_len}} "
            "({count:>{count_len}} elements)".format(
                label=label,
                count=count,
                percentage=count / float(total_count) * 100.0,
                label_len=label_len,
                count_len=count_len,
            )
        )


def get_distribution(labels):
    """
    Get the distribution of the labels.

    Prameters
    ---------
    labels : list of str
        This list is non-unique.

    Returns
    -------
    distribution : dict
        Maps (label => count)

    Examples
    --------
    >>> dist = get_distribution(['de', 'de', 'en'])
    >>> sorted(dist.items())
    [('de', 2), ('en', 1)]
    """
    distribution = {}
    for label in labels:
        if label not in distribution:
            distribution[label] = 1
        else:
            distribution[label] += 1
    return distribution



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Optimize confusion matrix.

For more information, see

* http://cs.stackexchange.com/q/70627/2914
* http://datascience.stackexchange.com/q/17079/8820
"""

# core modules
from pkg_resources import resource_filename
import json
import logging
import random

# 3rd party modules
from jinja2 import Template
from mpl_toolkits.axes_grid1 import make_axes_locatable
from sklearn.metrics import silhouette_score
import matplotlib.pyplot as plt
import numpy as np

# internal modules
import clana.clustering
import clana.cm_metrics
import clana.io
import clana.utils

cfg = clana.utils.load_cfg()


def main(
    cm_file,
    perm_file,
    steps,
    labels_file,
    zero_diagonal,
    limit_classes=None,
    output=None,
):
    """
    Run optimization and generate output.

    Parameters
    ----------
    cm_file : str
    perm_file : str
    steps : int
    labels_file : str
    zero_diagonal : bool
    limit_classes : int, optional (default: no limit)
    output : str
    """
    cm = clana.io.read_confusion_matrix(cm_file)
    perm = clana.io.read_permutation(cm_file, perm_file)
    labels = clana.io.read_labels(labels_file, len(cm))
    n, m = cm.shape
    if n != m:
        raise ValueError(
            "Confusion matrix is expected to be square, but was {} x {}".format(n, m)
        )
    if len(labels) - 1 != n:
        print(
            "Confusion matrix is {n} x {n}, but len(labels)={nb_labels}".format(
                n=n, nb_labels=len(labels)
            )
        )

    cm_orig = cm.copy()

    get_cm_problems(cm, labels)

    weights = calculate_weight_matrix(len(cm))
    print("Score: {}".format(calculate_score(cm, weights)))
    result = simulated_annealing(
        cm, perm, score=calculate_score, deterministic=True, steps=steps
    )
    print("Score: {}".format(calculate_score(result["cm"], weights)))
    print("Perm: {}".format(list(result["perm"])))
    clana.io.ClanaCfg.store_permutation(cm_file, result["perm"], steps)
    labels = [labels[i] for i in result["perm"]]
    class_indices = list(range(len(labels)))
    class_indices = [class_indices[i] for i in result["perm"]]
    logging.info("Classes: {}".format(labels))
    acc = clana.cm_metrics.get_accuracy(cm_orig)
    print("Accuracy: {:0.2f}%".format(acc * 100))
    start = 0
    if limit_classes is None:
        limit_classes = len(cm)
    if output is None:
        output = cfg["visualize"]["save_path"]
    plot_cm(
        result["cm"][start:limit_classes, start:limit_classes],
        zero_diagonal=zero_diagonal,
        labels=labels[start:limit_classes],
        output=output,
    )
    create_html_cm(
        result["cm"][start:limit_classes, start:limit_classes],
        zero_diagonal=zero_diagonal,
        labels=labels[start:limit_classes],
    )
    if len(cm) < 5:
        print(
            "You only have {} classes. Clustering for less than 5 classes "
            "should be done manually.".format(len(cm))
        )
        return
    grouping = clana.clustering.extract_clusters(result["cm"], labels)
    y_pred = [0]
    cluster_i = 0
    for el in grouping:
        if el == 1:
            cluster_i += 1
        y_pred.append(cluster_i)
    logging.info("silhouette_score={}".format(silhouette_score(cm, y_pred)))
    # Store grouping as hierarchy
    with open(cfg["visualize"]["hierarchy_path"], "w") as outfile:
        hierarchy = clana.clustering.apply_grouping(class_indices, grouping)
        hierarchy = clana.clustering._remove_single_element_groups(hierarchy)
        str_ = json.dumps(
            hierarchy,
            indent=4,
            sort_keys=True,
            separators=(",", ":"),
            ensure_ascii=False,
        )
        outfile.write(str_)

    # Print nice
    for group in clana.clustering.apply_grouping(labels, grouping):
        print(u"\t{}: {}".format(len(group), [el for el in group]))


def get_cm_problems(cm, labels):
    """
    Find problems of a classifier by analzing its confusion matrix.

    Parameters
    ----------
    cm : ndarray
    labels : List[str]
    """
    n = len(cm)

    # Find classes which are not present in the dataset
    for i in range(n):
        if sum(cm[i]) == 0:
            logging.warning("The class '{}' was not in the dataset.".format(labels[i]))

    # Find classes which are never predicted
    cm = cm.transpose()
    never_predicted = []
    for i in range(n):
        if sum(cm[i]) == 0:
            never_predicted.append(labels[i])
    if len(never_predicted) > 0:
        logging.warning(
            "The following classes were never predicted: {}".format(never_predicted)
        )


def calculate_score(cm, weights):
    """
    Calculate a score how close big elements of cm are to the diagonal.

    Examples
    --------
    >>> cm = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    >>> weights = calculate_weight_matrix(3)
    >>> weights.shape
    (3, 3)
    >>> calculate_score(cm, weights)
    32
    """
    return int(np.tensordot(cm, weights, axes=((0, 1), (0, 1))))


def calculate_weight_matrix(n):
    """
    Calculate the weights for each position.

    The weight is the distance to the diagonal.

    Parameters
    ----------
    n : int

    Examples
    --------
    >>> calculate_weight_matrix(3)
    array([[0.  , 1.01, 2.02],
           [1.01, 0.  , 1.03],
           [2.02, 1.03, 0.  ]])
    """
    weights = np.abs(np.arange(n) - np.arange(n)[:, None])
    weights = np.array(weights, dtype=np.float)
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            weights[i][j] += (i + j) * 0.01
    return weights


def swap(cm, i, j):
    """
    Swap row and column i and j in-place.

    Parameters
    ----------
    cm : ndarray
    i : int
    j : int

    Examples
    --------
    >>> cm = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    >>> swap(cm, 2, 0)
    array([[8, 7, 6],
           [5, 4, 3],
           [2, 1, 0]])
    """
    # swap columns
    copy = cm[:, i].copy()
    cm[:, i] = cm[:, j]
    cm[:, j] = copy
    # swap rows
    copy = cm[i, :].copy()
    cm[i, :] = cm[j, :]
    cm[j, :] = copy
    return cm


def move_1d(perm, from_start, from_end, insert_pos):
    """
    Move a block in a list.

    Parameters
    ----------
    perm : ndarray
        Permutation
    from_start : int
    from_end : int
    insert_pos : int

    Returns
    -------
    perm : ndarray
        The new permutation
    """
    assert insert_pos < from_start or insert_pos > from_end
    if insert_pos > from_end:
        p_new = list(range(from_end + 1, insert_pos + 1)) + list(
            range(from_start, from_end + 1)
        )
    else:
        p_new = list(range(from_start, from_end + 1)) + list(
            range(insert_pos, from_start)
        )
    p_old = sorted(p_new)
    perm[p_old] = perm[p_new]
    return perm


def move(cm, from_start, from_end, insert_pos):
    """
    Move rows from_start - from_end to insert_pos in-place.

    Parameters
    ----------
    cm : ndarray
    from_start : int
    from_end : int
    insert_pos : int

    Returns
    -------
    cm : ndarray

    Examples
    --------
    >>> cm = np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 0, 1], [2, 3, 4, 5]])
    >>> move(cm, 1, 2, 0)
    array([[5, 6, 4, 7],
           [9, 0, 8, 1],
           [1, 2, 0, 3],
           [3, 4, 2, 5]])
    """
    assert insert_pos < from_start or insert_pos > from_end
    if insert_pos > from_end:
        p_new = list(range(from_end + 1, insert_pos + 1)) + list(
            range(from_start, from_end + 1)
        )
    else:
        p_new = list(range(from_start, from_end + 1)) + list(
            range(insert_pos, from_start)
        )
    p_old = sorted(p_new)
    # swap columns
    cm[:, p_old] = cm[:, p_new]
    # swap rows
    cm[p_old, :] = cm[p_new, :]
    return cm


def swap_1d(perm, i, j):
    """
    Swap two elements of a 1-D numpy array in-place.

    Parameters
    ----------
    parm : ndarray
    i : int
    j : int

    Examples
    --------
    >>> perm = np.array([2, 1, 2, 3, 4, 5, 6])
    >>> swap_1d(perm, 2, 6)
    array([2, 1, 6, 3, 4, 5, 2])
    """
    perm[i], perm[j] = perm[j], perm[i]
    return perm


def apply_permutation(cm, perm):
    """
    Apply permutation to a matrix.

    Parameters
    ----------
    cm : ndarray
    perm : List[int]

    Examples
    --------
    >>> cm = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    >>> perm = np.array([2, 0, 1])
    >>> apply_permutation(cm, perm)
    array([[8, 6, 7],
           [2, 0, 1],
           [5, 3, 4]])
    """
    return cm[perm].transpose()[perm].transpose()


def simulated_annealing(
    current_cm,
    current_perm=None,
    score=calculate_score,
    steps=2 * 10 ** 5,
    temp=100.0,
    cooling_factor=0.99,
    deterministic=False,
):
    """
    Optimize current_cm by randomly swapping elements.

    Parameters
    ----------
    current_cm : ndarray
    current_perm : None or iterable, optional (default: None)
    steps : int, optional (default: 2 * 10**4)
    temp : float > 0.0, optional (default: 100.0)
        Temperature
    cooling_factor: float in (0, 1), optional (default: 0.99)

    Returns
    -------
    best_result : Dict[str, Any]
        "best_cm"
        "best_perm"
    """
    assert temp > 0.0
    assert cooling_factor > 0.0
    assert cooling_factor < 1.0
    n = len(current_cm)
    logging.info("n={}".format(n))

    # Load the initial permutation
    if current_perm is None:
        current_perm = list(range(n))
    current_perm = np.array(current_perm)

    # Pre-calculate weights
    weights = calculate_weight_matrix(n)

    # Apply the permutation
    current_cm = apply_permutation(current_cm, current_perm)
    current_score = score(current_cm, weights)

    best_cm = current_cm
    best_score = current_score
    best_perm = current_perm

    logging.info("## Starting Score: {:0.2f}".format(current_score))
    for step in range(steps):
        tmp_cm = np.array(current_cm, copy=True)

        swap_prob = 0.5
        make_swap = random.random() < swap_prob
        if n < 3:
            # In this case block-swaps don't make any sense
            make_swap = True
        if make_swap:
            # Choose what to swap
            i = random.randint(0, n - 1)
            j = i
            while j == i:
                j = random.randint(0, n - 1)
            # Define permutation
            perm = swap_1d(current_perm.copy(), i, j)
            # Define values after swap
            tmp_cm = swap(tmp_cm, i, j)
        else:
            # block-swap
            block_len = n
            while block_len >= n - 1:
                from_start = random.randint(0, n - 3)
                from_end = random.randint(from_start + 1, n - 2)
                block_len = from_start - from_end
            insert_pos = from_start
            while not (insert_pos < from_start or insert_pos > from_end):
                insert_pos = random.randint(0, n - 1)
            perm = move_1d(current_perm.copy(), from_start, from_end, insert_pos)

            # Define values after swap
            tmp_cm = move(tmp_cm, from_start, from_end, insert_pos)
        tmp_score = score(tmp_cm, weights)

        # Should be swapped?
        if deterministic:
            chance = 1.0
        else:
            chance = random.random()
            temp *= 0.99
        hot_prob_thresh = min(1, np.exp(-(tmp_score - current_score) / temp))
        if chance <= hot_prob_thresh:
            changed = False
            if best_score > tmp_score:  # minimize
                best_perm = perm
                best_cm = tmp_cm
                best_score = tmp_score
                changed = True
            current_score = tmp_score
            current_cm = tmp_cm
            current_perm = perm
            if changed:
                logging.info(
                    (
                        "Current: %0.2f (best: %0.2f, "
                        "hot_prob_thresh=%0.4f%%, step=%i, swap=%s)"
                    ),
                    current_score,
                    best_score,
                    (hot_prob_thresh * 100),
                    step,
                    str(make_swap),
                )
    return {"cm": best_cm, "perm": best_perm}


def plot_cm(cm, zero_diagonal=False, labels=None, output=cfg["visualize"]["save_path"]):
    """
    Plot a confusion matrix.

    Parameters
    ----------
    cm : ndarray
    zero_diagonal : bool, optional (default: False)
    labels : list of str, optional
        If this is not given, then numbers are assigned to the classes
    """
    from matplotlib.colors import LogNorm

    n = len(cm)
    if zero_diagonal:
        for i in range(n):
            cm[i][i] = 0
    if n > 20:
        size = int(n / 4.0)
    else:
        size = 5
    fig = plt.figure(figsize=(size, size), dpi=80)
    plt.clf()
    ax = fig.add_subplot(111)
    ax.set_aspect(1)
    if labels is None:
        labels = [i for i in range(len(cm))]
    x = [i for i in range(len(cm))]
    plt.xticks(x, labels, rotation=cfg["visualize"]["xlabels_rotation"])
    y = [i for i in range(len(cm))]
    plt.yticks(y, labels, rotation=cfg["visualize"]["ylabels_rotation"])
    if cfg["visualize"]["norm"] == "LogNorm":
        norm = LogNorm(vmin=max(1, np.min(cm)), vmax=np.max(cm))
    elif cfg["visualize"]["norm"] is None:
        norm = None
    else:
        raise NotImplementedError(
            "visualize->norm={} is not implemented. " "Try None or LogNorm"
        )
    res = ax.imshow(
        np.array(cm),
        cmap=cfg["visualize"]["colormap"],
        interpolation=cfg["visualize"]["interpolation"],
        norm=norm,
    )
    width, height = cm.shape

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.5)
    plt.colorbar(res, cax=cax)
    plt.tight_layout()

    logging.info("Save figure at '{}'".format(output))
    plt.savefig(output)


def create_html_cm(cm, zero_diagonal=False, labels=None):
    """
    Plot a confusion matrix.

    Parameters
    ----------
    cm : ndarray
    zero_diagonal : bool, optional (default: False)
        If this is set to True, then the diagonal is overwritten with zeroes.
    labels : list of str, optional
        If this is not given, then numbers are assigned to the classes
    """
    if labels is None:
        labels = [i for i in range(len(cm))]

    el_max = 200

    template_path = resource_filename("clana", "templates/base.html")
    with open(template_path, "r") as f:
        base = f.read()

    cm_t = cm.transpose()
    header_cells = []
    for i, label in enumerate(labels):
        precision = cm[i][i] / float(sum(cm_t[i]))
        background_color = "transparent"
        if precision < 0.2:
            background_color = "red"
        elif precision > 0.98:
            background_color = "green"
        header_cells.append(
            {
                "precision": "{:0.2f}".format(precision),
                "background-color": background_color,
                "label": label,
            }
        )

    body_rows = []
    for i, label, row in zip(range(len(labels)), labels, cm):
        body_row = []
        row_str = [str(el) for el in row]
        support = sum(row)
        recall = cm[i][i] / float(support)
        background_color = "transparent"
        if recall < 0.2:
            background_color = "red"
        elif recall >= 0.98:
            background_color = "green"
        body_row.append(
            {
                "label": label,
                "recall": "{:.2f}".format(recall),
                "background-color": background_color,
            }
        )
        for j, pred_label, el in zip(range(len(labels)), labels, row_str):
            background_color = "transparent"
            if el == "0":
                el = ""
            else:
                background_color = get_color_code(float(el), el_max)

            body_row.append(
                {
                    "label": el,
                    "true": label,
                    "pred": pred_label,
                    "background-color": background_color,
                }
            )

        body_rows.append({"row": body_row, "support": support})

    html = Template(base)
    html = html.render(header_cells=header_cells, body_rows=body_rows)

    with open(cfg["visualize"]["html_save_path"], "w") as f:
        f.write(html)


def get_color(white_to_black):
    """
    Get grayscale color.

    Parameters
    ----------
    white_to_black : float

    Returns
    -------
    color : tuple

    Examples
    --------
    >>> get_color(0)
    (255, 255, 255)
    >>> get_color(0.5)
    (128, 128, 128)
    >>> get_color(1)
    (0, 0, 0)
    """
    assert 0 <= white_to_black <= 1
    # in HSV, red is 0 deg and green is 120 deg (out of 360);
    # divide red_to_green with 3 to map [0, 1] to [0, 1./3.]
    # hue = red_to_green / 3.0
    # r, g, b = colorsys.hsv_to_rgb(hue, 0, 1)
    # return map(lambda x: int(255 * x), (r, g, b))

    index = 255 - int(255 * white_to_black)
    r, g, b = index, index, index
    return int(r), int(g), int(b)


def get_color_code(val, max_val):
    """
    Get a HTML color code which is between 0 and max_val.

    Parameters
    ----------
    val : number
    max_val : number

    Returns
    -------
    color_code : str

    Examples
    --------
    >>> get_color_code(0, 100)
    '#ffffff'
    >>> get_color_code(100, 100)
    '#000000'
    >>> get_color_code(50, 100)
    '#808080'
    """
    value = min(1.0, float(val) / max_val)
    r, g, b = get_color(value)
    return "#{:02x}{:02x}{:02x}".format(r, g, b)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Calculate the confusion matrix (CSV inputs)."""

# core modules
import csv
import logging

# 3rd party modules
import numpy as np

# internal modules
import clana.io


def main(cm_dump_filepath, gt_filepath, n):
    """
    Calculate a confusion matrix.

    Parameters
    ----------
    cm_dump_filepath : str
        CSV file with delimter ; and quoting char "
        The first field is an identifier, the second one is the index of the
        predicted label
    gt_filepath : str
        CSV file with delimter ; and quoting char "
        The first field is an identifier, the second one is the index of the
        ground truth
    n : int
        Number of classes
    """
    cm = calculate_cm(cm_dump_filepath, gt_filepath, n)
    path = "cm.json"
    clana.io.write_cm(path, cm)
    logging.info("cm was written to '{}'".format(path))


def calculate_cm(cm_dump_filepath, gt_filepath, n):
    """
    Calculate a confusion matrix.

    Parameters
    ----------
    cm_dump_filepath : str
        CSV file with delimter ; and quoting char "
        The first field is an identifier, the second one is the index of the
        predicted label
    gt_filepath : str
        CSV file with delimter ; and quoting char "
        The first field is an identifier, the second one is the index of the
        ground truth
    n : int
        Number of classes

    Returns
    -------
    confusion_matrix : numpy array (n x n)
    """
    cm = np.zeros((n, n), dtype=int)

    # Read CSV files
    with open(cm_dump_filepath, "r") as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        predictions = [row for row in reader]

    with open(gt_filepath, "r") as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        truths = [row for row in reader]

    ident2truth_index = {}
    for identifier, truth_index in truths:
        ident2truth_index[identifier] = int(truth_index)

    if len(predictions) != len(truths):
        msg = 'len(predictions) = {} != {} = len(truths)"'.format(
            len(predictions), len(truths)
        )
        raise ValueError(msg)

    for ident, pred_index in predictions:
        cm[ident2truth_index[ident]][int(pred_index)] += 1

    return cm



def get_accuracy(cm):
    """
    Get the accuaracy by the confusion matrix cm.

    Parameters
    ----------
    cm : ndarray

    Returns
    -------
    accuracy : float

    Examples
    --------
    >>> import numpy as np
    >>> cm = np.array([[10, 20], [30, 40]])
    >>> get_accuracy(cm)
    0.5
    >>> cm = np.array([[20, 10], [30, 40]])
    >>> get_accuracy(cm)
    0.6
    """
    return float(sum([cm[i][i] for i in range(len(cm))])) / float(cm.sum())



# core modules
import csv
import hashlib
import json
import os

# 3rd party modules
import numpy as np
import yaml

# internal modules
import clana.utils


class ClanaCfg(object):
    @classmethod
    def read_clana_cfg(cls, cfg_file):
        """
        Read a .clana config file which contains permutations.

        Parameters
        ----------
        cfg_file : str

        Returns
        -------
        cfg : Dict[str, Any]
        """
        if os.path.isfile(cfg_file):
            with open(cfg_file, "r") as stream:
                cfg = yaml.safe_load(stream)
        else:
            cfg = {"version": clana.__version__, "data": {}}
        return cfg

    @classmethod
    def get_cfg_path_from_cm_path(cls, cm_file):
        return os.path.join(os.path.dirname(os.path.abspath(cm_file)), ".clana")

    @classmethod
    def get_perm(cls, cm_file):
        """
        Get the best permutation found so far for a given cm_file.

        Fallback: list(range(n))

        Parameters
        ----------
        cm_file : str

        Returns
        -------
        perm : List[int]
        """
        cfg_file = cls.get_cfg_path_from_cm_path(cm_file)
        cfg = cls.read_clana_cfg(cfg_file)
        cm_file_base = os.path.basename(cm_file)
        cm = read_confusion_matrix(cm_file)
        n = len(cm)
        perm = list(range(n))
        if cm_file_base in cfg["data"]:
            cm_file_md5 = md5(cm_file)
            if cm_file_md5 in cfg["data"][cm_file_base]:
                print(
                    "Loaded permutation found in {} iterations".format(
                        cfg["data"][cm_file_base][cm_file_md5]["iterations"]
                    )
                )
                perm = cfg["data"][cm_file_base][cm_file_md5]["permutation"]
        return perm

    @classmethod
    def store_permutation(cls, cm_file, permutation, iterations):
        """
        Store a permutation.

        Parameters
        ----------
        cm_file : str
        permutation : List[int]
        iterations : int
        """
        cm_file = os.path.abspath(cm_file)
        cfg_file = cls.get_cfg_path_from_cm_path(cm_file)
        if os.path.isfile(cfg_file):
            cfg = ClanaCfg.read_clana_cfg(cfg_file)
        else:
            cfg = {"version": clana.__version__, "data": {}}

        cm_file_base = os.path.basename(cm_file)
        if cm_file_base not in cfg["data"]:
            cfg["data"][cm_file_base] = {}
        cm_file_md5 = md5(cm_file)
        if cm_file_md5 not in cfg["data"][cm_file_base]:
            cfg["data"][cm_file_base][cm_file_md5] = {
                "permutation": permutation.tolist(),
                "iterations": 0,
            }
        cfg["data"][cm_file_base][cm_file_md5]["permutation"] = permutation.tolist()
        cfg["data"][cm_file_base][cm_file_md5]["iterations"] += iterations

        # Write file
        print(cfg_file)
        with open(cfg_file, "w") as outfile:
            yaml.dump(cfg, outfile, default_flow_style=False, allow_unicode=True)


def read_confusion_matrix(cm_file, make_max=float("inf")):
    """
    Load confusion matrix.

    Parameters
    ----------
    cm_file : str
        Path to a JSON file which contains a confusion matrix (List[List[int]])
    make_max : float, optional (default: +Infinity)
        Crop values at this value.

    Returns
    -------
    cm : np.array
    """
    with open(cm_file) as f:
        if cm_file.lower().endswith("csv"):
            cm = []
            with open(cm_file, newline="") as csvfile:
                spamreader = csv.reader(csvfile, delimiter=",", quotechar='"')
                for row in spamreader:
                    cm.append([int(el) for el in row])
        else:
            cm = json.load(f)
        cm = np.array(cm)

    # Crop values
    n = len(cm)
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            cm[i][j] = min(cm[i][j], make_max)

    return cm


def read_permutation(cm_file, perm_file):
    """
    Load permutation.

    Parameters
    ----------
    cm_file : str
    perm_file : str or None
        Path to a JSON file which contains a permutation of n numbers.

    Returns
    -------
    perm : List[int]
        Permutation of the numbers 0, ..., n-1
    """
    assert os.path.isfile(cm_file)
    cm = read_confusion_matrix(cm_file)
    n = len(cm)
    if perm_file is not None and os.path.isfile(perm_file):
        with open(perm_file) as data_file:
            if perm_file.lower().endswith("csv"):
                with open(perm_file) as file:
                    content = file.read()
                perm = [int(el) for el in content.split(",")]
            else:
                perm = json.load(data_file)
    else:
        perm = ClanaCfg.get_perm(cm_file)
    return perm


def read_labels(labels_file, n):
    """
    Load labels.

    Please note that this contains one additional "UNK" label for
    unknown classes.

    Parameters
    ----------
    labels_file : str
    n : int

    Returns
    -------
    labels : List[str]
    """
    labels = clana.utils.load_labels(labels_file, n)
    labels.append("UNK")
    return labels


def write_labels(labels_file, labels):
    with open(labels_file, "w") as outfile:
        str_ = json.dumps(labels, indent=2, separators=(",", ": "), ensure_ascii=False)
        outfile.write(str_)


def write_predictions(identifier2prediction, filepath):
    """
    Create a predictions file.

    Parameters
    ----------
    identifier2prediction : Dict[str, str]
        Map an identifier (as used in write_gt) to a prediction.
        The prediction is a single class, not a distribution.
    filepath : str
        Write to this CSV file.
    """
    with open(filepath, "w") as f:
        for identifier, prediction in identifier2prediction.items():
            f.write("{};{}\n".format(identifier, prediction))


def write_gt(identifier2label, filepath):
    """
    Write ground truth to a file.

    Parameters
    ----------
    identifier2label : Dict[str, str]
    filepath : str
        Write to this CSV file.
    """
    with open(filepath, "w") as f:
        for identifier, label in identifier2label.items():
            f.write("{};{}\n".format(identifier, label))


def write_cm(path, cm):
    """
    Write confusion matrix to path.

    Parameters
    ----------
    path : str
    cm : ndarray
    """
    with open(path, "w") as outfile:
        str_ = json.dumps(cm.tolist(), separators=(",", ": "), ensure_ascii=False)
        outfile.write(str_)


def md5(fname):
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Calculate the confusion matrix (one label per line)."""

# core modules
import csv
import json
import logging
import os
import sys

# 3rd party modules
import numpy as np
import sklearn.metrics

# internal modules
import clana.utils


def main(label_filepath, gt_filepath, predictions_filepath, clean):
    """
    Get a simple confunsion matrix.

    Parameters
    ----------
    label_filepath : str
        Path to a CSV file with delimiter ;
    gt_filepath : str
        Path to a CSV file with delimiter ;
    predictions : str
        Path to a CSV file with delimiter ;
    clean : bool, optional (default: False)
        Remove classes that the classifier doesn't know
    """
    label_filepath = os.path.abspath(label_filepath)
    labels = clana.utils.load_labels(label_filepath, 0)

    # Read CSV files
    with open(gt_filepath, "r") as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        truths = [row[0] for row in reader]

    with open(predictions_filepath, "r") as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        predictions = [row[0] for row in reader]

    cm = calculate_cm(labels, truths, predictions, clean=False)
    # Write JSON file
    cm_filepath = os.path.abspath("cm.json")
    logging.info("Write results to '{}'.".format(cm_filepath))
    with open(cm_filepath, "w") as outfile:
        str_ = json.dumps(
            cm.tolist(), indent=2, separators=(",", ": "), ensure_ascii=False
        )
        outfile.write(str_)
    print(cm)


def calculate_cm(labels, truths, predictions, replace_unk_preds=False, clean=False):
    """
    Calculate a confusion matrix.

    Parameters
    ----------
    labels : list
    truths : list
    predictions : list
    replace_unk_preds : bool, optional (default: True)
        If a prediction is not in the labels in label_filepath, replace it
        with UNK
    clean : bool, optional (default: False)
        Remove classes that the classifier doesn't know

    Returns
    -------
    confusion_matrix : numpy array (n x n)
    """
    # Check data
    if len(predictions) != len(truths):
        msg = 'len(predictions) = {} != {} = len(truths)"'.format(
            len(predictions), len(truths)
        )
        raise ValueError(msg)

    label2i = {}  # map a label to 0, ..., n
    for i, label in enumerate(labels):
        label2i[label] = i

    if clean:
        logging.debug("@" * 80)
        preds = []
        truths_tmp = []
        for tru, pred in zip(truths, predictions):
            if tru in predictions:
                truths_tmp.append(tru)
                preds.append(pred)
        predictions = preds
        truths = truths_tmp

    if replace_unk_preds:
        preds = []
        for pred in predictions:
            if label in label2i:
                preds.append(label)
            else:
                preds.append("UNK")
        predictions = preds

    # Sanity check
    for label in truths:
        if label not in label2i:
            logging.error("Could not find label '{}'".format(label))
            sys.exit(-1)

    n = len(labels)
    for label in predictions:
        if label not in label2i:
            label2i[label] = len(labels)
            n = len(labels) + 1
            logging.error(
                "Could not find label '{}' in labels file => "
                "Add class UNK".format(label)
            )

    # TODO: do no always filter
    filter_data_unk = True
    if filter_data_unk:
        truths2, predictions2 = [], []
        for tru, pred in zip(truths, predictions):
            if pred != "unk":  # TODO: tru != 'UNK'!!!
                truths2.append(tru)
                predictions2.append(pred)
        truths = truths2
        predictions = predictions2

    report = sklearn.metrics.classification_report(truths, predictions, labels=labels)
    print(report)
    print(
        "Accuracy: {:.2f}%".format(
            sklearn.metrics.accuracy_score(truths, predictions) * 100
        )
    )

    cm = np.zeros((n, n), dtype=int)

    for truth_label, pred_label in zip(truths, predictions):
        cm[label2i[truth_label]][label2i[pred_label]] += 1

    return cm



# Core modules
import logging
import random

# 3rd party modules
import numpy as np

# internal modules
import clana.utils

cfg = clana.utils.load_cfg()


def apply_grouping(labels, grouping):
    """
    Return list of grouped labels.

    Parameters
    ----------
    labels : list
    grouping : list of bool

    Examples
    --------
    >>> labels = ['de', 'en', 'fr']
    >>> grouping = [False, True]
    >>> apply_grouping(labels, grouping)
    [['de', 'en'], ['fr']]
    """
    groups = []
    current_group = [labels[0]]
    for label, cut in zip(labels[1:], grouping):
        if cut:
            groups.append(current_group)
            current_group = [label]
        else:
            current_group.append(label)
    groups.append(current_group)
    return groups


def _remove_single_element_groups(hierarchy):
    """
    Flatten sub-lists of length 1.

    Parameters
    ----------
    hierarchy : list of lists

    Returns
    -------
    hierarchy : list of el / lists

    Examples
    --------
    >>> hierarchy = [[0], [1, 2]]
    >>> _remove_single_element_groups(hierarchy)
    [0, [1, 2]]
    """
    h_new = []
    for el in hierarchy:
        if len(el) > 1:
            h_new.append(el)
        else:
            h_new.append(el[0])
    return h_new


def extract_clusters(
    cm,
    labels,
    steps=10 ** 4,
    lambda_=0.013,
    method="local-connectivity",
    interactive=False,
):
    """
    Find clusters in cm.

    Idea:
        mininmize lambda (error between clusters) - (count of clusters)
        s.t.: Each inter-cluster accuracy has to be lower than the overall
              accuracy

    Parameters
    ----------
    cm : ndarray
    labels : list
    steps : int
    lambda_ : float
        The closer to 0, the more groups
        The bigger, the bigger groups
    method : {'local-connectivity', 'energy'}
    interactive : bool

    Returns
    -------
    clustes : list of lists of labels
    """

    def create_weight_matrix(grouping):
        n = len(grouping) + 1
        weight_matrix = np.zeros((n, n))
        for i in range(n):
            seen_1 = False
            for j in range(i + 1, n):
                if seen_1:
                    weight_matrix[i][j] = 1
                elif grouping[j - 1] == 1:
                    seen_1 = True
                    weight_matrix[i][j] = 1
        return weight_matrix + weight_matrix.transpose()

    def get_score(cm, grouping, lambda_):
        from clana.visualize_cm import calculate_score

        inter_cluster_err = 0.0
        weights = create_weight_matrix(grouping)
        inter_cluster_err = calculate_score(cm, weights)
        return lambda_ * inter_cluster_err - sum(grouping)

    def find_thres(cm, percentage):
        """
        Find a threshold for grouping.

        Parameters
        ----------
        cm : ndarray
        percentage : float
            Probability that two neighboring classes belong togehter
        """
        n = int(len(cm) * (1.0 - percentage)) - 1
        con = sorted(get_neighboring_connectivity(cm))
        return con[n]

    def find_thres_interactive(cm, labels):
        """
        Find a threshold for grouping.

        The threshold is the minimum connection strength for two classes to be
        within the same cluster.

        Parameters
        ----------
        cm : ndarray
        percentage : float
            Probability that two neighboring classes belong togehter
        """
        n = len(cm)
        con = sorted(
            zip(get_neighboring_connectivity(cm), zip(range(n - 1), range(1, n)))
        )
        # pos_low = 0
        pos_str = None

        # Lowest position from which we know that they are connected
        pos_up = n - 1

        # Highest position from which we know that they are not connected
        neg_low = 0
        # neg_up = n - 1
        while pos_up - 1 > neg_low:
            print("pos_up={}, neg_low={}, pos_str={}".format(pos_up, neg_low, pos_str))
            pos = int((pos_up + neg_low) / 2)
            con_str, (i1, i2) = con[pos]
            should_be_conn = raw_input(
                "Should {} and {} be in one cluster?"
                " (y/n): ".format(labels[i1], labels[i2])
            )
            if should_be_conn == "n":
                neg_low = pos
            elif should_be_conn == "y":
                pos_up = pos
                pos_str = con_str
            else:
                print(
                    "Please type only 'y' or 'n'. You typed {}.".format(should_be_conn)
                )
        return pos_str

    def get_neighboring_connectivity(cm):
        con = []
        n = len(cm)
        for i in range(n - 1):
            con.append(cm[i][i + 1] + cm[i + 1][i])
        return con

    def split_at_con_thres(cm, thres, labels, interactive):
        """
        Two classes are not in the same group if they are not connected strong.

        Minimum connection strength is thres. The bigger this value, the more
        clusters / the smaller clusters you will get.
        """
        con = get_neighboring_connectivity(cm)
        grouping = []
        for i, el in enumerate(con):
            if el == thres and interactive:
                should_conn = "-"
                while should_conn not in ["y", "n"]:
                    should_conn = raw_input(
                        "Should {} and {} be in one "
                        "cluster? (y/n): ".format(labels[i], labels[i + 1])
                    )
                    if should_conn == "y":
                        grouping.append(0)
                    elif should_conn == "n":
                        grouping.append(1)
                    else:
                        print("please type either 'y' or 'n'")
            else:
                grouping.append(el < thres)
        return grouping

    if method == "energy":
        n = len(cm)
        grouping = np.zeros(n - 1)
        minimal_score = get_score(cm, grouping, lambda_)
        best_grouping = grouping.copy()
        for i in range(steps):
            pos = random.randint(0, n - 2)
            grouping = best_grouping.copy()
            grouping[pos] = (grouping[pos] + 1) % 2
            current_score = get_score(cm, grouping, lambda_)
            if current_score < minimal_score:
                best_grouping = grouping
                minimal_score = current_score
                logging.info(
                    "Best grouping: {} (score: {})".format(grouping, minimal_score)
                )
    elif method == "local-connectivity":
        if interactive:
            thres = find_thres_interactive(cm, labels)
        else:
            thres = find_thres(cm, cfg["visualize"]["threshold"])
        logging.info("Found threshold for local connection: {}".format(thres))
        best_grouping = split_at_con_thres(cm, thres, labels, interactive=interactive)
    else:
        raise NotImplementedError("method='{}'".format(method))
    logging.info("Found {} clusters".format(sum(best_grouping) + 1))
    return best_grouping



"""CLANA is a toolkit for classifier analysis."""

from setuptools import find_packages
from setuptools import setup
import io
import os

here = os.path.abspath(os.path.dirname(__file__))
try:
    # obtain long description from README and CHANGES
    # Specify encoding to get a unicode type in Python 2 and a str in Python 3
    readme_path = os.path.join(here, "docs/README.md")
    with io.open(readme_path, "r", encoding="utf-8") as f:
        README = f.read()
except IOError:
    README = ""

requires_tests = [
    "pytest",
    "pytest-cov",
    "pytest-mccabe",
    "pytest-flake8",
    "pytest-black",
]

install_requires = [
    "click>=6.7",
    "jinja2",
    "matplotlib>=2.1.1",
    "numpy>=1.14.0",
    "PyYAML>=5.1.1",
    "scikit-learn>=0.19.1",
    "scipy>=1.0.0",
]


config = {
    "name": "clana",
    "version": "0.3.8",
    "author": "Martin Thoma",
    "author_email": "info@martin-thoma.de",
    "maintainer": "Martin Thoma",
    "maintainer_email": "info@martin-thoma.de",
    "packages": find_packages(),
    "entry_points": {"console_scripts": ["clana=clana.cli:entry_point"]},
    "platforms": ["Linux"],
    "url": "https://github.com/MartinThoma/clana",
    "license": "MIT",
    "description": __doc__,
    "long_description": README,
    "long_description_content_type": "text/markdown",
    "install_requires": install_requires,
    "tests_require": requires_tests,
    "keywords": ["Machine Learning", "Data Science"],
    "download_url": "https://github.com/MartinThoma/clana",
    "package_data": {"clana": ["config.yaml"]},
    "include_package_data": True,
    "classifiers": [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],
    "zip_safe": False,
}

setup(**config)
